import re
import sys
import os
import numpy as np
from numpy.polynomial import Polynomial

# Commanded speeds and actual measured speeds (edit these values as needed)
commanded_speeds = [8000, 12000, 18000, 24000]  # Commanded RPM values
actual_speeds = [7998, 11370, 16710, 24000]     # Measured RPM values

# The comment that will be added to the corrected file
correction_comment = "(C3D Spindle Error Corrected)"

def third_order_correction(speed, a, b, c, d):
    """
    Apply a third-order correction to the spindle speed (scaled by 1000).
    
    The correction function is: f(speed/1000) = a * (speed/1000)^3 + b * (speed/1000)^2 + c * (speed/1000) + d
    
    Parameters:
        speed (int or float): Original spindle speed.
        a, b, c, d (float): Coefficients for the third-order correction.
        
    Returns:
        float: Recalibrated spindle speed.
    """
    scaled_speed = speed / 1000  # Scale the speed by 1000
    return a * scaled_speed**3 + b * scaled_speed**2 + c * scaled_speed + d

def calculate_least_squares_coefficients(commanded_speeds, actual_speeds):
    """
    Perform least squares regression to find the cubic polynomial coefficients.
    
    Parameters:
        commanded_speeds (list): List of commanded speeds (RPM).
        actual_speeds (list): List of actual measured speeds (RPM).
        
    Returns:
        tuple: Coefficients (a, b, c, d) for the third-order correction function.
    """
    # Scale commanded speeds by dividing by 1000 for the regression
    actual_speeds_scaled = np.array(actual_speeds) / 1000.0
    commanded_speeds = np.array(commanded_speeds)

    # Perform least squares regression for a 3rd-degree polynomial fit
    p = Polynomial.fit(actual_speeds_scaled, commanded_speeds, 3)

    # Return the coefficients in the correct order: a, b, c, d
    return p.convert().coef[::-1]  # Convert to standard polynomial and reverse

def modify_gcode_spindle_speed(input_file, a, b, c, d):
    """
    Modify spindle speeds in G-Code file by applying a third-order correction function.
    
    Parameters:
        input_file (str): Path to the input G-Code file.
        a, b, c, d (float): Coefficients for the third-order correction function.
        
    Output:
        A new G-Code file with a "-fixed" suffix on the filename.
    """
    # Create the output file name by appending "-fixed" before the file extension
    file_root, file_ext = os.path.splitext(input_file)
    output_file = f"{file_root}-fixed{file_ext}"
    
    with open(input_file, 'r') as infile, open(output_file, 'w') as outfile:
        # First, check if the file has already been corrected
        for line in infile:
            if correction_comment in line:
                print(f"Error: The file '{input_file}' has already been corrected.")
                outfile.close()  # Ensure the output file is closed
                os.remove(output_file)  # Delete the empty output file
                sys.exit(1)
        
        # Reset file pointer after checking
        infile.seek(0)

        # Write the correction comment at the top of the output file
        outfile.write(correction_comment + '\n')

        # Process the G-Code file line by line
        for line in infile:
            # Search for lines containing the spindle speed with M3 (SXXXM3)
            match = re.search(r'(S)(\d+)(M3)', line)
            if match:
                original_speed = int(match.group(2))  # Extract the spindle speed
                # Apply the third-order correction function
                corrected_speed = third_order_correction(original_speed, a, b, c, d)
                # Replace the old speed with the new corrected speed in the same format
                corrected_line = f"S{int(corrected_speed)}M3"
                outfile.write(corrected_line + '\n')
            else:
                # Write the line unchanged if it doesn't contain an SXXXM3 command
                outfile.write(line)
    
    print(f"Modified G-Code file created: {output_file}")

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python modify_gcode.py <input_gcode_file>")
        sys.exit(1)

    input_file = sys.argv[1]

    # Calculate the least squares regression coefficients
    a, b, c, d = calculate_least_squares_coefficients(commanded_speeds, actual_speeds)
    
    print(f"Calculated coefficients: a = {a}, b = {b}, c = {c}, d = {d}")

    # Call the function to modify the G-Code file
    modify_gcode_spindle_speed(input_file, a, b, c, d)
